//
//  tank
//
//  Import and Playlist Daemon for autoradio project
//
//
//  Copyright (C) 2017-2018 Christian Pointner <equinox@helsinki.at>
//
//  This file is part of tank.
//
//  tank is free software: you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation, either version 3 of the License, or
//  any later version.
//
//  tank is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with tank. If not, see <http://www.gnu.org/licenses/>.
//

package importer

import (
	"fmt"
	"log"
	"net/http"
	"strings"
	"time"

	"gitlab.servus.at/autoradio/tank/store"
)

type newSessionResponse struct {
	id           uint64
	session      *Session
	responsecode int
	errorstring  string
}

type newSessionRequest struct {
	ctx      *ImportContext
	refId    string
	response chan newSessionResponse
}

type getSessionResponse struct {
	session      *Session
	refId        string
	responsecode int
	errorstring  string
}

type getSessionRequest struct {
	group    string
	id       uint64
	response chan getSessionResponse
}

type SessionsUpdateCB func(added, removed map[uint64]string, userdata interface{}) bool

type SessionsListCB struct {
	cb       SessionsUpdateCB
	userdata interface{}
}

type listSessionsResponse struct {
	sessions     map[uint64]string
	responsecode int
	errorstring  string
}

type listSessionsRequest struct {
	group    string
	callback SessionsUpdateCB
	userdata interface{}
	response chan listSessionsResponse
}

type removeSessionResponse struct {
	responsecode int
	errorstring  string
}

type removeSessionRequest struct {
	group    string
	id       uint64
	response chan removeSessionResponse
}

type inventorySession struct {
	s     *session
	refId string
}

type inventoryGroup struct {
	sessions  map[uint64]*inventorySession
	updateCBs []SessionsListCB
}

func (group *inventoryGroup) callUpdateHandler(added, removed map[uint64]string) {
	var keptCBs []SessionsListCB
	for _, cb := range group.updateCBs {
		if cb.cb != nil {
			if keep := cb.cb(added, removed, cb.userdata); keep {
				keptCBs = append(keptCBs, cb)
			}
		}
	}
	group.updateCBs = keptCBs
}

func (group *inventoryGroup) callUpdateHandlerAdd(id uint64, refId string) {
	added := make(map[uint64]string)
	added[id] = refId
	group.callUpdateHandler(added, nil)
}

func (group *inventoryGroup) callUpdateHandlerRemove(id uint64, refId string) {
	removed := make(map[uint64]string)
	removed[id] = refId
	group.callUpdateHandler(nil, removed)
}

type SessionInventory struct {
	groups     map[string]*inventoryGroup
	conf       *Config
	store      *store.Store
	infoLog    *log.Logger
	errLog     *log.Logger
	dbgLog     *log.Logger
	quit       chan bool
	done       chan bool
	newChan    chan newSessionRequest
	getChan    chan getSessionRequest
	listChan   chan listSessionsRequest
	removeChan chan removeSessionRequest
	workerChan chan<- workerRequest
}

func (inventory *SessionInventory) new(ctx *ImportContext, refId string) (resp newSessionResponse) {
	resp.responsecode = http.StatusOK
	resp.errorstring = "OK"

	id, err := inventory.store.CreateFile(ctx.GroupName, *store.NewFile(ctx.UserName))
	if err != nil {
		resp.responsecode = http.StatusInternalServerError
		resp.errorstring = err.Error()
		return
	}

	resp.id = id
	if _, exists := inventory.groups[ctx.GroupName]; !exists {
		newgroup := &inventoryGroup{}
		newgroup.sessions = make(map[uint64]*inventorySession)
		inventory.groups[ctx.GroupName] = newgroup
	}
	ctx.conf = inventory.conf
	if pref := ctx.infoLog.Prefix(); strings.Contains(pref, "%s") {
		ctx.infoLog.SetPrefix(fmt.Sprintf(pref, resp.id))
	}
	if pref := ctx.errLog.Prefix(); strings.Contains(pref, "%s") {
		ctx.errLog.SetPrefix(fmt.Sprintf(pref, resp.id))
	}
	if pref := ctx.dbgLog.Prefix(); strings.Contains(pref, "%s") {
		ctx.dbgLog.SetPrefix(fmt.Sprintf(pref, resp.id))
	}

	s, err := newSession(ctx, inventory.workerChan, resp.id, func() { inventory.Remove(ctx.GroupName, resp.id) })
	if err != nil {
		resp.responsecode = http.StatusInternalServerError
		resp.errorstring = err.Error()
		return
	}
	inventory.groups[ctx.GroupName].sessions[resp.id] = &inventorySession{s, refId}
	resp.session = inventory.groups[ctx.GroupName].sessions[resp.id].s.public()
	inventory.dbgLog.Printf("importer: created session for '%s' -> %d", ctx.GroupName, resp.id)
	inventory.groups[ctx.GroupName].callUpdateHandlerAdd(resp.id, refId)
	return
}

func (inventory *SessionInventory) get(groupname string, id uint64) (resp getSessionResponse) {
	resp.responsecode = http.StatusOK
	resp.errorstring = "OK"

	group, exists := inventory.groups[groupname]
	if !exists {
		resp.responsecode = http.StatusNotFound
		resp.errorstring = fmt.Sprintf("importer: session '%s/%d' not found", groupname, id)
		return
	}

	if session, exists := group.sessions[id]; exists {
		resp.session = session.s.public()
		resp.refId = session.refId
		return
	}
	resp.responsecode = http.StatusNotFound
	resp.errorstring = fmt.Sprintf("importer: session '%s/%d' not found", groupname, id)
	return
}

func (inventory *SessionInventory) list(groupname string, userdata interface{}, cb SessionsUpdateCB) (resp listSessionsResponse) {
	resp.responsecode = http.StatusOK
	resp.errorstring = "OK"
	resp.sessions = make(map[uint64]string)
	if group, exists := inventory.groups[groupname]; exists {
		for id, e := range group.sessions {
			resp.sessions[id] = e.refId
		}
		if cb != nil {
			group.updateCBs = append(group.updateCBs, SessionsListCB{cb, userdata})
		}
	} else if cb != nil {
		newgroup := &inventoryGroup{}
		newgroup.sessions = make(map[uint64]*inventorySession)
		newgroup.updateCBs = []SessionsListCB{SessionsListCB{cb, userdata}}
		inventory.groups[groupname] = newgroup
	}
	return
}

func (inventory *SessionInventory) remove(groupname string, id uint64) (resp removeSessionResponse) {
	resp.responsecode = http.StatusOK
	resp.errorstring = "OK"

	group, exists := inventory.groups[groupname]
	if !exists {
		resp.responsecode = http.StatusNotFound
		resp.errorstring = fmt.Sprintf("importer: session '%s/%d' not found", groupname, id)
		return
	}

	if session, exists := group.sessions[id]; exists {
		go session.s.cleanup() // cleanup could take a while -> don't block all the other stuff
		refId := session.refId
		delete(group.sessions, id)
		inventory.dbgLog.Printf("importer: removed session '%s/%d'", groupname, id)
		group.callUpdateHandlerRemove(id, refId)

		if len(group.sessions) == 0 && len(group.updateCBs) == 0 {
			delete(inventory.groups, groupname)
			inventory.dbgLog.Printf("importer: removed group '%s'", groupname)
		}
	} else {
		resp.responsecode = http.StatusNotFound
		resp.errorstring = fmt.Sprintf("importer: session '%s/%d' not found", groupname, id)
	}
	return
}

func (inventory *SessionInventory) maintenanceTask() {
	for name, group := range inventory.groups {
		group.callUpdateHandler(nil, nil)
		if len(group.sessions) == 0 && len(group.updateCBs) == 0 {
			delete(inventory.groups, name)
			inventory.dbgLog.Printf("importer: removed group '%s'", name)
		}
	}
}

func (inventory *SessionInventory) dispatchRequests() {
	defer func() { inventory.done <- true }()

	inventory.dbgLog.Println("importer: session-inventory initialized")

	mt := time.NewTicker(1 * time.Minute)
	for {
		select {
		case <-inventory.quit:
			return
		case <-mt.C:
			inventory.maintenanceTask()
		case req := <-inventory.newChan:
			req.response <- inventory.new(req.ctx, req.refId)
		case req := <-inventory.getChan:
			req.response <- inventory.get(req.group, req.id)
		case req := <-inventory.listChan:
			req.response <- inventory.list(req.group, req.userdata, req.callback)
		case req := <-inventory.removeChan:
			req.response <- inventory.remove(req.group, req.id)
		}
	}
}

// *********************************************************
// Public Interface

func (inventory *SessionInventory) New(ctx *ImportContext, refId string) (uint64, *Session, int, string) {
	resCh := make(chan newSessionResponse)
	req := newSessionRequest{}
	req.ctx = ctx
	req.refId = refId
	req.response = resCh
	inventory.newChan <- req

	res := <-resCh
	return res.id, res.session, res.responsecode, res.errorstring
}

func (inventory *SessionInventory) Get(group string, id uint64) (*Session, string, int, string) {
	resCh := make(chan getSessionResponse)
	req := getSessionRequest{}
	req.group = group
	req.id = id
	req.response = resCh
	inventory.getChan <- req

	res := <-resCh
	return res.session, res.refId, res.responsecode, res.errorstring
}

func (inventory *SessionInventory) List(group string, userdata interface{}, cb SessionsUpdateCB) (map[uint64]string, int, string) {
	resCh := make(chan listSessionsResponse)
	req := listSessionsRequest{}
	req.group = group
	req.response = resCh
	req.callback = cb
	req.userdata = userdata
	inventory.listChan <- req

	res := <-resCh
	return res.sessions, res.responsecode, res.errorstring
}

func (inventory *SessionInventory) Remove(group string, id uint64) (int, string) {
	resCh := make(chan removeSessionResponse)
	req := removeSessionRequest{}
	req.group = group
	req.id = id
	req.response = resCh
	inventory.removeChan <- req

	res := <-resCh
	return res.responsecode, res.errorstring
}

func (inventory *SessionInventory) Cleanup() {
	inventory.quit <- true
	<-inventory.done
	close(inventory.quit)
	close(inventory.done)
	close(inventory.newChan)
	close(inventory.getChan)
	close(inventory.listChan)
	close(inventory.removeChan)
}

func newSessionInventory(conf *Config, st *store.Store, infoLog, errLog, dbgLog *log.Logger, workerChan chan<- workerRequest) (inventory *SessionInventory, err error) {
	inventory = &SessionInventory{}
	inventory.conf = conf
	inventory.store = st
	inventory.infoLog = infoLog
	inventory.errLog = errLog
	inventory.dbgLog = dbgLog
	inventory.quit = make(chan bool, 1)
	inventory.done = make(chan bool)
	inventory.groups = make(map[string]*inventoryGroup)
	inventory.newChan = make(chan newSessionRequest, 10)
	inventory.getChan = make(chan getSessionRequest, 10)
	inventory.listChan = make(chan listSessionsRequest, 10)
	inventory.removeChan = make(chan removeSessionRequest, 10)
	inventory.workerChan = workerChan

	go inventory.dispatchRequests()
	return
}
